import numpy as np
import numpy.testing as npt
import pytest

import pynbody


def setup_module():
    # create a dummy gadget file
    global f, h
    f = pynbody.new(dm=2097152)
    f['iord'] = np.arange(2097152)
    f.properties['z']=1.6591479493605812
    f._filename = "testdata/rockstar/snapshot_015"
    with pytest.warns(RuntimeWarning, match=r"No iord array available; assuming.*"):
        h = f.halos()


def test_load_rockstar():
    global f, h
    assert len(h)==5851
    assert isinstance(h, pynbody.halo.RockstarCatalogue)

def test_rockstar_properties():
    global h
    h_properties = h[4977].properties
    assert h_properties['num_p']==40
    npt.assert_allclose(h_properties['pos'], [43.892704, 0.197397, 40.751919], rtol=1e-6)

def test_rockstar_particles():
    global h
    assert (h[4977]['iord']==[1801964, 1802346, 1818475, 1818729, 1818730, 1818857, 1818858, 1818859, 1818986,
                             1834860, 1834986, 1834987, 1835113, 1835114, 1835115, 1835116, 1835242, 1835243,
                             1835244, 1835369, 1835370, 1835371, 1835498, 1835499, 1851372, 1851625, 1851626,
                             1851627, 1851628, 1851754, 1851755, 1851756, 1851884, 1868010, 1868011, 1868012,
                             1884394, 1884395, 1900651, 1933291]).all()

def test_reject_unsuitable_rockstar_files():
    fwrong = pynbody.new(dm=2097152)
    fwrong.properties['z']=0
    with pytest.raises(RuntimeError):
        hwrong = fwrong.halos()



import glob
import os.path
import shutil
import stat
import subprocess

import numpy.testing as npt
import pytest

import pynbody


def test_load_ahf_catalogue():
    f = pynbody.load("testdata/g15784.lr.01024")
    h = pynbody.halo.AHFCatalogue(f)
    assert len(h)==1411

def test_load_ahf_catalogue_non_gzipped():
    subprocess.call(["gunzip","testdata/g15784.lr.01024.z0.000.AHF_halos.gz"])
    try:
        f = pynbody.load("testdata/g15784.lr.01024")
        h = pynbody.halo.AHFCatalogue(f)
        assert len(h)==1411
    finally:
        subprocess.call(["gzip","testdata/g15784.lr.01024.z0.000.AHF_halos"])

def test_ahf_properties():
    f = pynbody.load("testdata/g15784.lr.01024")
    h = pynbody.halo.AHFCatalogue(f)
    assert h[1].properties['children']==[]
    assert h[1].properties['fstart']==23


@pytest.fixture
def setup_unwritable_ahf_situation():
    if os.path.exists("testdata/test_unwritable"):
        os.chmod("testdata/test_unwritable", stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
        shutil.rmtree("testdata/test_unwritable")
    os.mkdir("testdata/test_unwritable/")
    for fname in glob.glob("testdata/g15784*"):
        if "AHF_fpos" not in fname:
            os.symlink("../"+fname[9:], "testdata/test_unwritable/"+fname[9:])
    os.chmod("testdata/test_unwritable", stat.S_IRUSR | stat.S_IXUSR)


def test_ahf_unwritable(setup_unwritable_ahf_situation):
    f = pynbody.load("testdata/test_unwritable/g15784.lr.01024")
    with pytest.warns(UserWarning, match="Unable to write AHF_fpos file;.*"):
         h = f.halos()
    assert len(h)==1411


def test_detecting_ahf_catalogues_with_without_trailing_slash():
    # Test small fixes in #688 to detect AHF catalogues with and wihtout trailing slashes in directories
    for name in (
        "testdata/ramses_new_format_cosmo_with_ahf_output_00110",
        "testdata/ramses_new_format_cosmo_with_ahf_output_00110/"
    ):
        f = pynbody.load(name)
        _halos = pynbody.halo.AHFCatalogue(f)


def test_ramses_ahf_family_mapping_with_new_format():
    # Test Issue 691 where family mapping of AHF catalogues with Ramses new particle formats would go wrong
    f = pynbody.load("testdata/ramses_new_format_cosmo_with_ahf_output_00110")
    halos = pynbody.halo.AHFCatalogue(f)

    assert len(halos) == 149    # 150 lines in AHF halos file

    # Load halos and check that stars, DM and gas are correctly mapped by pynbody
    # Halo 1 is the main halo and has all three families, while other are random picks
    halo_numbers = [1, 10, 15]
    for halo_number in halo_numbers:
        halo = halos[halo_number]

        # There should not be any extra families in the halo particles
        assert(all(fam in [pynbody.family.dm, pynbody.family.star, pynbody.family.gas] for fam in halo.families()))

        # Check we now have the same number of particles assigned to the halo
        # than its AHF header, family by family
        assert halo.properties['npart'] == len(halo)
        assert halo.properties['n_star'] == len(halo.st)
        assert halo.properties['n_gas'] == len(halo.g)
        ndm = halo.properties['npart'] - halo.properties['n_star'] - halo.properties['n_gas']
        assert ndm == len(halo.d)

        # Derive some masses to check that we are identifying the right particles, in addition to their right numbers
        dm_mass = halo.properties['Mhalo'] - halo.properties['M_star'] - halo.properties['M_gas']
        gas_mass = halo.properties['M_gas']

        rtol = 1e-2 # We are not precise to per cent level with unit conversion through the different steps
        hubble = f.properties['h']      # AHF internal units are Msol/h and need to be manually corrected, which has not been done on this test output
        npt.assert_allclose(dm_mass / hubble, halo.d['mass'].sum().in_units("Msol"), rtol=rtol)
        npt.assert_allclose(gas_mass / hubble, halo.g['mass'].sum().in_units("Msol"), rtol=rtol)
        npt.assert_allclose(halo.properties['Mhalo'] / hubble, halo['mass'].sum().in_units("Msol"), rtol=rtol)



import os.path

import numpy as np

import pynbody.snapshot.simsnap

from . import Halo, HaloCatalogue


class RockstarIntermediateCatalogue(HaloCatalogue):
    """Reader for Rockstar intermediate catalogues as generated by
    Michael Tremmel's tool"""

    _halo_type = np.dtype([('id',np.int64),('num_p',np.int64),('indstart',np.int64)])
    _part_type = np.dtype('int64')

    def __init__(self, sim, sort=True, correct=False, **kwargs):
        assert isinstance(sim, pynbody.snapshot.simsnap.SimSnap)
        self._correct=correct
        HaloCatalogue.__init__(self, sim)

        self._halos = {}

        self._index_filename = sim.filename+".rockstar.halos"
        self._particles_filename = sim.filename+".rockstar.halo_particles_fpos"

        self._init_index()
        if sort:
            self._sort_index()

    def __len__(self):
        return self._nhalos

    def _get_particles_for_halo(self, num):
        halo_info = self._halo_info[num]
        with util.open_(self._particles_filename) as f:
            f.seek(halo_info['indstart']*self._part_type.itemsize)
            halo_ptcls=np.fromfile(f,dtype=self._part_type,count=halo_info['num_p']-1)
            halo_ptcls.sort()


        return halo_ptcls

    def _add_halo_id(self, halo, num):
        halo.properties['rockstar_halo_id']=self._halo_info[num]['id']
        if self._correct is True:
            with util.open_(self._particles_filename) as f:
                f.seek(self._halo_info['indstart'][num]*self._part_type.itemsize)
                one_ptcl = np.fromfile(f,dtype=self._part_type,count=1)
                from . import load
                onep = load(self.base.filename, take = one_ptcl)
                while not onep['grpid']:
                    one_ptcl = np.fromfile(f,dtype=self._part_type,count=1)
                    onep = load(self.base.filename, take = one_ptcl)
                halo.properties['rockstar_halo_id'] = onep['grpid'][0]

    def _get_halo(self, i):
        if self.base is None:
            raise RuntimeError("Parent SimSnap has been deleted")

        halo_ptcls = self._get_particles_for_halo(i)
        h = Halo(i, self, self.base, halo_ptcls)
        self._add_halo_id(h,i)
        return h

    def load_copy(self, i):
        """Load a fresh SimSnap with only the particles in halo i"""
        if i>=len(self):
            raise KeyError("No such halo")

        from . import load
        halo = load(self.base.filename, take=self._get_particles_for_halo(i))
        self._add_halo_id(halo, i)
        return halo

    @staticmethod
    def _can_load(sim,*args,**kwargs):
        return os.path.exists(sim.filename+".rockstar.halos") and \
               os.path.exists(sim.filename+".rockstar.halo_particles_fpos")

    def _init_index(self):
        with open(self._index_filename,"rb") as f:
            self._nhalos = np.fromfile(f,np.int64,1)[0]
            #self._halo_info = np.fromfile(f,self._halo_type,self._nhalos)
            self._halo_info = np.fromfile(f,self._halo_type,-1)
            ok, = np.where(self._halo_info['indstart']>=0)
            self._halo_info = self._halo_info[ok]

    def _sort_index(self):
        self._halo_info[::-1].sort(order='num_p')

    def get_group_array(self, family='star'):
        if family == 'star':
            target = self.base.star
        if family == 'gas':
            target = self.base.gas
        if family == 'dark':
            target = self.base.dark
        if family == 'BH':
            temptarget = self.base.star
            target = temptarget[(temptarget['tform']<0)]
        if family not in ['star','gas','dark', 'BH']:
            raise TypeError("family input not understood. Must be star, gas, or dark")
        ar = -1 * np.ones(len(target))
        if family != 'dark':
            fpos_ar = target.get_index_list(self.base)
        ngas = len(self.base.gas)
        ndark = len(self.base.dark)
        with util.open_(self._particles_filename) as f:
            for i in range(len(self._halo_info)):
                #print i, len(self._halo_info)
                #if i%100 == 0: print float(i)/float(len(self._halo_info))*100, "% done"
                f.seek(self._halo_info[i]['indstart']*self._part_type.itemsize)
                halo_ptcls=np.fromfile(f,dtype=self._part_type,count=self._halo_info[i]['num_p']-1)
                if family == 'gas':
                    halo_ptcls = halo_ptcls[(halo_ptcls>=ndark)&(halo_ptcls<ndark+ngas)]
                if family == 'dark':
                    halo_ptcls = halo_ptcls[(halo_ptcls<ndark)]
                if family == 'star' or family == 'BH':
                    halo_ptcls = halo_ptcls[(halo_ptcls>=ndark+ngas)]
                if family != 'dark':
                    match, = np.where(np.in1d(fpos_ar, halo_ptcls))
                    ar[match] = i
                else:
                    ar[halo_ptcls] = i
        return ar.astype(np.int_)



import os.path
import re
from typing import Sequence

import numpy as np

from .. import array, units, util
from ..extern.cython_fortran_utils import FortranFile
from . import DummyHalo, Halo, HaloCatalogue, logger

unit_length = units.Unit("Mpc")
unit_vel = units.Unit("km s**-1")
unit_mass = 1e11 * units.Unit("Msol")
unit_angular_momentum = unit_mass * unit_vel * unit_length
unit_energy = unit_mass * unit_vel ** 2
unit_temperature = units.Unit("K")
unit_density = unit_mass / unit_length ** 3

UNITS = {
    'pos_x': unit_length,
    'pos_y': unit_length,
    'pos_z': unit_length,
    'shape_a': unit_length,
    'shape_b': unit_length,
    'shape_c': unit_length,
    'nfw_R_c': unit_length,
    'max_distance': unit_length,
    'R200c': unit_length,
    'r_half_mass': unit_length,
    'r_90percent_mass': unit_length,
    'max_velocity_radius': unit_length,
    'radius_profile': unit_length,
    'virial_radius': unit_length,
    'velocity_dispersion': unit_vel,
    'vel_x': unit_vel,
    'vel_y': unit_vel,
    'vel_z': unit_vel,
    'max_velocity': unit_vel,
    'virial_velocity': unit_vel,
    'angular_momentum_x': unit_angular_momentum,
    'angular_momentum_y': unit_angular_momentum,
    'angular_momentum_z': unit_angular_momentum,
    'm': unit_mass,
    'M200c': unit_mass,
    'm_contam': unit_mass,
    'mtot': unit_mass,
    'mtot_contam': unit_mass,
    'virial_mass': unit_mass,
    'kinetic_energy': unit_energy,
    'potential_energy': unit_energy,
    'total_energy': unit_energy,
    'virial_temperature': unit_temperature,
    'nfw_rho0': unit_density,
    'density_profile': unit_density,
}


class BaseAdaptaHOPCatalogue(HaloCatalogue):
    """A AdaptaHOP Catalogue. AdaptaHOP output files must be in
    Halos/<simulation_number>/ directory or specified by fname"""

    _AdaptaHOP_fname = None
    _halos = None
    _headers = None
    _fname = ""
    _read_contamination = False

    _halo_attributes = tuple()
    _halo_attributes_contam = tuple()
    _header_attributes = tuple()

    def __init__(self, sim, fname=None, read_contamination=None, longint=None):
        if FortranFile is None:
            raise RuntimeError(
                "Support for AdaptaHOP requires the package `cython-fortran-file` to be installed."
            )

        if fname is None:
            for fname in AdaptaHOPCatalogue._enumerate_hop_tag_locations_from_sim(sim):
                if os.path.exists(fname):
                    break

            if not os.path.exists(fname):
                raise RuntimeError(
                    "Unable to find AdaptaHOP brick file in simulation directory"
                )
        if read_contamination is None or longint is None:
            read_contamination, longint = self._detect_file_format(fname)
        self._read_contamination = read_contamination
        self._longint = longint

        self._header_attributes = self.convert_i8b(self._header_attributes, longint)
        self._halo_attributes = self.convert_i8b(self._halo_attributes, longint)
        self._halo_attributes_contam = self.convert_i8b(self._halo_attributes_contam, longint)

        self._header_attributes = self.convert_i8b(self._header_attributes, longint)
        self._halo_attributes = self.convert_i8b(self._halo_attributes, longint)
        self._halo_attributes_contam = self.convert_i8b(self._halo_attributes_contam, longint)

        # Call parent class
        super().__init__(sim)

        # Initialize internal data
        self._base_dm = sim.dm

        self._halos = {}
        self._fname = fname
        self._AdaptaHOP_fname = fname

        logger.debug("AdaptaHOPCatalogue loading offsets")

        # Compute offsets
        self._ahop_compute_offset()
        logger.debug("AdaptaHOPCatalogue loaded")

    def _detect_file_format(self, fname):
        """
        Detect if the file is in the old format or the new format.
        """
        with FortranFile(fname) as fpu:
            longint_flag = self._detect_longint(fpu, (False, True))

            # Now attempts reading the first halo data
            attrs, attrs_contam = (self.convert_i8b(_, longint_flag) for _ in (self._halo_attributes, self._halo_attributes_contam))

            if len(attrs_contam) == 0:
                read_contamination = False
            else:
                fpu.skip(3) # number + ids of parts + halo_ID
                fpu.read_attrs(attrs)
                try:
                    fpu.read_attrs(attrs_contam)
                    read_contamination = True
                except (ValueError, OSError):
                    read_contamination = False

        return read_contamination, longint_flag

    @staticmethod
    def convert_i8b(original_headers, longint):
        headers = []
        for key, count, dtype in original_headers:
            if dtype == "i8b":
                if longint:
                    dtype = "l"
                else:
                    dtype = "i"
            headers.append((key, count, dtype))
        return tuple(headers)

    def precalculate(self):
        """Speed up future operations by precalculating the indices
        for all halos in one operation. This is slow compared to
        getting a single halo, however."""
        # Get the mapping from particle to halo
        self._base_dm._family_index()  # filling the cache
        self._group_array = self.get_group_array(group_to_indices=True)

    def _ahop_compute_offset(self):
        """
        Compute the offset in the brick file of each halo.
        """

        with FortranFile(self._fname) as fpu:
            self._headers = fpu.read_attrs(self._header_attributes)

            nhalos = self._headers["nhalos"]
            nsubs = self._headers["nsubs"]

            Nskip = len(self._halo_attributes)
            if self._read_contamination:
                Nskip += len(self._halo_attributes_contam)

            for _ in range(nhalos + nsubs):
                ipos = fpu.tell()
                fpu.skip(2)  # number + ids of parts
                halo_ID = fpu.read_int()
                fpu.skip(Nskip)

                # Fill-in data
                dummy = DummyHalo()
                dummy.properties["file_offset"] = ipos
                self._halos[halo_ID] = dummy

    def calc_item(self, halo_id):
        return self._get_halo(halo_id)

    def _get_halo(self, halo_id):
        if halo_id not in self._halos:
            raise KeyError(f"Halo with id '{halo_id}' does not seem to exist in the catalog.")

        halo = self._halos[halo_id]
        halo_dummy = self._halos[halo_id]
        if isinstance(halo, DummyHalo):
            halo = self._read_halo_data(halo_id, halo.properties["file_offset"])
            halo.dummy = halo_dummy
            self._halos[halo_id] = halo

        return halo

    def _read_member_helper(self, fpu, expected_size):
        """Read the member array from file, and return it *sorted*.

        The function automatically find whether the particle ids are stored in
        32 or 64 bits.
        """
        default_dtype = getattr(self.base, "_iord_dtype", "i")
        possible_dtypes = list({"i", "q"} - {default_dtype})

        dtypes = [default_dtype] + possible_dtypes
        ipos = fpu.tell()

        for dtype in dtypes:
            try:
                iord_array = fpu.read_vector(dtype)
                if iord_array.size == expected_size:
                    # Store dtype for next time
                    self.base._iord_dtype = dtype

                    if not util.is_sorted(iord_array) == 1:
                        return np.sort(iord_array)
                    else:
                        return iord_array

            except ValueError:
                pass
            # Rewind
            fpu.seek(ipos)

        # Could not read, throw an error
        raise RuntimeError("Could not read iord!")

    def _read_halo_data(self, halo_id, offset):
        """
        Reads the halo data from file -- AdaptaHOP specific.

        Parameters
        ----------
        halo_id : int
            The id of the halo
        offset : int
            The location in the file (in bytes)

        Returns
        -------
        halo : Halo object
            The halo object, filled with the data read from file.
        """
        with FortranFile(self._fname) as fpu:
            fpu.seek(offset)
            if self._longint:
                npart = fpu.read_int64()
            else:
                npart = fpu.read_int()
            iord_array = self._read_member_helper(fpu, npart)
            halo_id_read = fpu.read_int()
            assert halo_id == halo_id_read
            if self._read_contamination:
                attrs = self._halo_attributes + self._halo_attributes_contam
            else:
                attrs = self._halo_attributes
            props = fpu.read_attrs(attrs)

        # /!\: AdaptaHOP assumes that 1Mpc == 3.08e24 (exactly)
        boxsize = self.base.properties["boxsize"]
        Mpc2boxsize = boxsize.in_units("cm") / 3.08e24  # Hard-coded in AdaptaHOP...

        # Add units for known fields
        # NOTE: we need to list the items, as the dictionary is updated in place
        for k, v in list(props.items()):
            if k in ("pos_x", "pos_y", "pos_z"):
                # convert positions between [-Lbox/2, Lbox/2] to [0, Lbox].
                v = v / Mpc2boxsize + 0.5
                unit = boxsize
            elif k in UNITS:
                unit = UNITS[k]
            else:
                continue
            props[k] = array.SimArray(
                v,
                sim=self.base,
                units=unit,
                dtype=v.dtype,
            )

        props["file_offset"] = offset
        props["npart"] = npart
        props["members"] = iord_array
        # Create position and velocity
        props["pos"] = array.SimArray(
            [props["pos_x"], props["pos_y"], props["pos_z"]],
            units=props["pos_x"].units,
            sim=self.base,
        )
        props["vel"] = array.SimArray(
            [props["vel_x"], props["vel_y"], props["vel_z"]],
            units=props["vel_x"].units,
            sim=self.base,
        )

        # Create halo object and fill properties
        if hasattr(self, "_group_to_indices"):
            index_array = self._group_to_indices[halo_id]
            iord_array = None
        else:
            index_array = None
            iord_array = iord_array
        halo = Halo(
            halo_id, self, self._base_dm, index_array=index_array, iord_array=iord_array
        )
        for k, v in props.items():
            halo.properties[k] = v

        # Need to convert the units of the halo object as we
        # just updated them
        halo._autoconvert_properties()

        return halo

    def get_group_array(self, family="dm", group_to_indices=False):
        """Return an array with an integer for each particle in the simulation
        indicating which halo that particle is associated with. If there are multiple
        levels (i.e. subhalos), the number returned corresponds to the lowest level, i.e.
        the smallest subhalo.

        Arguments
        ---------
        family : optional, default : dm
            The family of the particles that make the group
        group_to_indices : optional, bool
            If True, store the mapping from groups to particle on-disk location.

        Returns
        -------
        igrp : int array
            An array that contains the index of the group that contains each particle.
        """
        logger.debug("Get_group_array")
        if family is None:
            family == self.base.families()[0]
        elif isinstance(family, str):
            families = self.base.families()
            matched_families = [f for f in families if f.name == family]
            if len(matched_families) != 1:
                raise Exception("Could not find family %s" % family)
            family = matched_families[0]
        try:
            data = self.base[family]
        except Exception:
            logger.error(type(self.base))
            logger.error(type(family))
            raise

        iord = data["iord"]
        iord_argsort = data["iord_argsort"]

        igrp = np.zeros(len(data), dtype=int) - 1

        if group_to_indices:
            grp2indices = {}
        with FortranFile(self._fname) as fpu:
            for halo_id, halo in self._halos.items():
                fpu.seek(halo.properties["file_offset"])
                if self._longint:
                    npart = fpu.read_int64()
                else:
                    npart = fpu.read_int()
                particle_ids = self._read_member_helper(fpu, npart)

                indices = util.binary_search(particle_ids, iord, iord_argsort)
                assert all(indices < len(iord))

                igrp[indices] = halo_id

                if group_to_indices:
                    grp2indices[halo_id] = indices
        if group_to_indices:
            self._group_to_indices = grp2indices
        return igrp

    @classmethod
    def _detect_longint(cls, fpu: FortranFile, longint_flags: Sequence) -> bool:
        for longint_flag in longint_flags:
            try:
                fpu.seek(0)
                fpu.read_attrs(cls.convert_i8b(cls._header_attributes, longint_flag))
                return longint_flag
            except (ValueError, OSError):
                pass

        raise ValueError("Could not detect longint")

    @classmethod
    def _can_load(cls, sim, arr_name="grp", *args, **kwa):
        candidates = [
            fname
            for fname in cls._enumerate_hop_tag_locations_from_sim(sim)
        ]
        valid_candidates = [fname for fname in candidates if os.path.exists(fname)]
        if len(valid_candidates) == 0:
            return False

        # Logic is as follows:
        # - If `longint` is provided, try loading with it.
        # - Otherwise, try loading with and without it
        use_longint = kwa.pop("longint", None)
        if use_longint is None:
            longint_flags = [True, False]
        else:
            longint_flags = use_longint
        for fname in valid_candidates:
            with FortranFile(fname) as fpu:
                try:
                    cls._detect_longint(fpu, longint_flags)
                    return True
                except ValueError:
                    pass
        return False

    @staticmethod
    def _enumerate_hop_tag_locations_from_sim(sim):
        try:
            match = re.search("output_([0-9]{5})", sim.filename)
            if match is None:
                raise OSError(
                    "Cannot guess the HOP catalogue filename for %s" % sim.filename
                )
            isim = int(match.group(1))
            name = "tree_bricks%03d" % isim
        except (OSError, ValueError):
            return []

        s_filename = os.path.abspath(sim.filename)
        s_dir = os.path.dirname(s_filename)

        ret = [
            os.path.join(s_filename, name),
            os.path.join(s_filename, "Halos", name),
            os.path.join(s_dir, "Halos", name),
            os.path.join(s_dir, "Halos", "%d" % isim, name),
        ]
        return ret

    def _can_run(self, *args, **kwa):
        return False


class NewAdaptaHOPCatalogue(BaseAdaptaHOPCatalogue):
    _header_attributes = (
        ("npart", 1, "i8b"),
        ("massp", 1, "d"),
        ("aexp", 1, "d"),
        ("omega_t", 1, "d"),
        ("age", 1, "d"),
        (("nhalos", "nsubs"), 2, "i"),
    )

    # List of the attributes read from file. This does *not* include the number of particles,
    # the list of particles, the id of the halo and the "timestep" (first 4 records).
    _halo_attributes = (
        ("timestep", 1, "i"),
        (
            ("level", "host_id", "first_subhalo_id", "n_subhalos", "next_subhalo_id"),
            5,
            "i",
        ),
        ("m", 1, "d"),
        ("ntot", 1, "i8b"),
        ("mtot", 1, "d"),
        # Note: we use pos_z instead of z to prevent confusion with redshift
        (("pos_x", "pos_y", "pos_z"), 3, "d"),
        (("vel_x", "vel_y", "vel_z"), 3, "d"),
        (("angular_momentum_x", "angular_momentum_y", "angular_momentum_z"), 3, "d"),
        (("max_distance", "shape_a", "shape_b", "shape_c"), 4, "d"),
        (("kinetic_energy", "potential_energy", "total_energy"), 3, "d"),
        ("spin", 1, "d"),
        ("velocity_dispersion", 1, "d"),
        (("virial_radius", "virial_mass", "virial_temperature", "virial_velocity"), 4, "d"),
        (("max_velocity_radius", "max_velocity"), 2, "d"),
        ("nfw_concentration", 1, "d"),
        (("R200c", "M200c"), 2, "d"),
        (("r_half_mass", "r_90percent_mass"), 2, "d"),
        ("radius_profile", -1, "d"),
        ("density_profile", -1, "d"),
        (("nfw_rho0", "nfw_R_c"), 2, "d"),
    )

    _halo_attributes_contam = (
        ("contaminated", 1, "i"),
        (("m_contam", "mtot_contam"), 2, "d"),
        (("n_contam", "ntot_contam"), 2, "i8b"),
    )


class AdaptaHOPCatalogue(BaseAdaptaHOPCatalogue):
    _header_attributes = (
        ("npart", 1, "i"),
        ("massp", 1, "f"),
        ("aexp", 1, "f"),
        ("omega_t", 1, "f"),
        ("age", 1, "f"),
        (("nhalos", "nsubs"), 2, "i"),
    )
    # List of the attributes read from file. This does *not* include the number of particles,
    # the list of particles, the id of the halo and the "timestep" (first 4 records).
    _halo_attributes = (
        ("timestep", 1, "i"),
        (
            ("level", "host_id", "first_subhalo_id", "n_subhalos", "next_subhalo_id"),
            5,
            "i",
        ),
        ("m", 1, "f"),
        # Note: we use pos_z instead of z to prevent confusion with redshift
        (("pos_x", "pos_y", "pos_z"), 3, "f"),
        (("vel_x", "vel_y", "vel_z"), 3, "f"),
        (("angular_momentum_x", "angular_momentum_y", "angular_momentum_z"), 3, "f"),
        (("max_distance", "shape_a", "shape_b", "shape_c"), 4, "f"),
        (("kinetic_energy", "potential_energy", "total_energy"), 3, "f"),
        ("spin", 1, "f"),
        (("virial_radius", "virial_mass", "virial_temperature", "virial_velocity"), 4, "f"),
        (("nfw_rho0", "nfw_R_c"), 2, "f"),
    )

    _halo_attributes_contam = tuple()



import glob
import gzip
import os.path
import re
import warnings

import numpy as np

from .. import config_parser, snapshot, util
from . import DummyHalo, Halo, HaloCatalogue, logger


class AHFCatalogue(HaloCatalogue):

    """
    Class to handle catalogues produced by Amiga Halo Finder (AHF).
    """

    def __init__(self, sim, make_grp=None, get_all_parts=None, use_iord=None, ahf_basename=None,
                 dosort=None, only_stat=None, write_fpos=True, **kwargs):
        """Initialize an AHFCatalogue.

        **kwargs** :

        *make_grp*: if True a 'grp' array is created in the underlying
                    snapshot specifying the lowest level halo that any
                    given particle belongs to. If it is False, no such
                    array is created; if None, the behaviour is
                    determined by the configuration system.

        *get_all_parts*: if True, the particle file is loaded for all halos.
                    Suggested to keep this None, as this is memory intensive.
                    The default function is to load in this data as needed.

        *use_iord*: if True, the particle IDs in the Amiga catalogue
                    are taken to refer to the iord array. If False,
                    they are the particle offsets within the file. If
                    None, the parameter defaults to True for
                    GadgetSnap, False otherwise.

        *ahf_basename*: specify the basename of the AHF halo catalog
                        files - the code will append 'halos',
                        'particles', and 'substructure' to this
                        basename to load the catalog data.

        *dosort*: specify if halo catalog should be sorted so that
                  halo 1 is the most massive halo, halo 2 the
                  second most massive and so on.

        *only_stat*: specify that you only wish to collect the halo
                    properties stored in the AHF_halos file and not
                    worry about particle information

        """

        import os.path
        if not self._can_load(sim, ahf_basename):
            self._run_ahf(sim)

        HaloCatalogue.__init__(self, sim)

        if use_iord is None:
            use_iord = isinstance(
                sim.ancestor,
                (snapshot.gadget.GadgetSnap, snapshot.gadgethdf.GadgetHDFSnap)
            )

        self._use_iord = use_iord

        self._all_parts = get_all_parts
        self._dosort = dosort
        self._only_stat = only_stat

        if ahf_basename is not None:
            self._ahfBasename = ahf_basename
        else:
            candidates = self._list_possible_candidates(sim, ahf_basename)
            if len(candidates) == 1:
                candidate = candidates[0]
            elif len(candidates) == 0:
                raise FileNotFoundError("No candidate AHF catalogue found; try specifying a catalogue using the "
                                        "ahf_basename keyword")
            else:
                candidate = candidates[0]
                warnings.warn(f"Multiple candidate AHF catalogues found; using {candidate}. To specify a different "
                              "catalogue, use the ahf_basename keyword, or move the other catalogues.")
            self._ahfBasename = util.cutgz(candidate)[:-9]

        try:
            with util.open_(self._ahfBasename + 'halos') as f:
                # The first line contains headers, need to skip it
                self._nhalos = sum(1 for i in f) - 1
        except OSError as e:
            raise FileNotFoundError(
                "Halo catalogue not found -- check the base name of catalogue "
                "data or try specifying a catalogue using the ahf_basename "
                "keyword"
            ) from e

        logger.info("AHFCatalogue loading particles")
        self._load_ahf_particles(self._ahfBasename + 'particles')

        logger.info("AHFCatalogue loading halos")
        self._load_ahf_halos(self._ahfBasename + 'halos')

        if self._only_stat is None:
            self._get_file_positions(self._ahfBasename + 'particles')

        if self._dosort is not None:
            nparr = np.array([self._halos[i+1].properties['npart'] for i in range(self._nhalos)])
            osort = np.argsort(nparr)[::-1]
            self._sorted_indices = osort + 1

        self._load_ahf_substructure(self._ahfBasename + 'substructure')

        if make_grp is None:
            make_grp = config_parser.getboolean('AHFCatalogue', 'AutoGrp')

        if make_grp:
            self.make_grp()

        if config_parser.getboolean('AHFCatalogue', 'AutoPid'):
            sim['pid'] = np.arange(0, len(sim))

        if write_fpos:
            if not os.path.exists(self._ahfBasename + 'fpos'):
                self._write_fpos()

        logger.info("AHFCatalogue loaded")

    def __getitem__(self,item):
        """
        get the appropriate halo if dosort is on
        """
        if self._dosort is not None:
            i = self._sorted_indices[item-1]
        else:
            i = item
        return super().__getitem__(i)

    def make_grp(self, name='grp'):
        """
        Creates a 'grp' array which labels each particle according to
        its parent halo.
        """
        self.base[name] = self.get_group_array()

    def _write_fpos(self):
        try:
            with open(self._ahfBasename + 'fpos', 'w') as f:
                for i in range(self._nhalos):
                    if i < self._nhalos - 1:
                        f.write(str(self._halos[i+1].properties['fstart'])+'\n')
                    else:
                        f.write(str(self._halos[i+1].properties['fstart']))
        except OSError:
            warnings.warn("Unable to write AHF_fpos file; performance will be reduced. Pass write_fpos=False to halo constructor to suppress this message.")

    def get_group_array(self, top_level=False, family=None):
        """
        output an array of group IDs for each particle.
        :top_level: If False, each particle associated with the lowest level halo they are in.
                    If True, each particle associated with their top-most level halo
        :family: specify the family of particles to output an array for (default is all particles)
        """

        target = None
        famslice = None

        if family is None:
            target = self.base
        else:
            if family in ['gas','star','dm']:
                famslice = self.base._get_family_slice(family)
                target = self.base[famslice]
            else:
                if family == 'bh':
                    temptarget = self.base.star
                    target = temptarget[(temptarget['tform']<0)]

        if target is None:
            raise ValueError("Family value given is not valid. Use 'gas', 'star', 'dm', or 'bh'")

        if self._dosort is None:
            #if we want to differentiate between top and bottom levels,
            #the halos do need to be in order regardless if dosort is on.
            nparr = np.array([self._halos[i+1].properties['npart'] for i in range(self._nhalos)])
            osort = np.argsort(nparr)[::-1]
            self._sorted_indices = osort + 1
            hcnt = self._sorted_indices

        else:
            hcnt = np.arange(len(self._sorted_indices)) + 1

        if top_level is False:
            hord = self._sorted_indices
        else:
            hord = self._sorted_indices[::-1]
            hcnt = hcnt[::-1]

        if self._all_parts is None:
            f = util.open_(self._ahfBasename+'particles')

        try:
            cnt = 0
            ar = np.full(len(target), -1, dtype=np.int32)

            for i in hord:
                halo = self._halos[i]
                if self._all_parts is not None:
                    ids = halo.get_index_list(self.base)
                else:
                    f.seek(halo.properties['fstart'], 0)
                    ids = self._load_ahf_particle_block(f, halo.properties['npart'])
                if family is None:
                    ar[ids] = hcnt[cnt]
                else:
                    if famslice:
                        t_mask = (ids >= famslice.start) & (ids < famslice.stop)
                        id_t = ids[np.where(t_mask)] - famslice.start
                    else:
                        fpos_ar = target.get_index_list(self.base)
                        id_t, = np.where(np.in1d(fpos_ar, ids))

                    ar[id_t] = hcnt[cnt]
                cnt += 1
        finally:
            if self._all_parts is None:
                f.close()
        return ar

    def _setup_children(self):
        """
        Creates a 'children' array inside each halo's 'properties'
        listing the halo IDs of its children. Used in case the reading
        of substructure data from the AHF-supplied _substructure file
        fails for some reason.
        """

        for i in range(self._nhalos):
            self._halos[i + 1].properties['children'] = []

        for i in range(self._nhalos):
            host = self._halos[i + 1].properties.get('hostHalo', -2)
            if host > -1:
                try:
                    self._halos[host + 1].properties['children'].append(i + 1)
                except KeyError:
                    pass

    def _get_halo(self, i):
        if self.base is None:
            raise RuntimeError("Parent SimSnap has been deleted")
        if self._all_parts is not None:
            return self._halos[i]

        with util.open_(self._ahfBasename+'particles') as f:
            fpos = self._halos[i].properties['fstart']
            f.seek(fpos,0)
            return Halo(
                i,
                self,
                self.base,
                self._load_ahf_particle_block(f, self._halos[i].properties['npart'])
            )



    @property
    def base(self):
        return self._base()

    def load_copy(self, i):
        """Load the a fresh SimSnap with only the particle in halo i"""

        from .. import load

        if self._dosort is not None:
            i = self._sorted_indices[i-1]

        with util.open_(self._ahfBasename + 'particles') as f:
            fpos = self._halos[i].properties['fstart']
            f.seek(fpos,0)
            ids = self._load_ahf_particle_block(f, nparts=self._halos[i].properties['npart'])

        return load(self.base.filename, take=ids)

    def _get_file_positions(self, filename):
        """Get the starting positions of each halo's particle information within the
        AHF_particles file for faster access later"""
        if os.path.exists(self._ahfBasename + 'fpos'):
            with util.open_(self._ahfBasename + 'fpos') as f:
                for i in range(self._nhalos):
                    self._halos[i+1].properties['fstart'] = int(f.readline())
        else:
            with util.open_(filename) as f:
                for h in range(self._nhalos):
                    if len(f.readline().split()) == 1:
                        f.readline()
                    self._halos[h+1].properties['fstart'] = f.tell()
                    for i in range(self._halos[h+1].properties['npart']):
                        f.readline()

    def _load_ahf_particle_block(self, f, nparts=None):
        """Load the particles for the next halo described in particle file f"""
        ng = len(self.base.gas)
        nd = len(self.base.dark)
        ns = len(self.base.star)
        nds = nd+ns

        if nparts is None:
            startline = f.readline()
            if len(startline.split())==1:
                startline = f.readline()
            nparts = int(startline.split()[0])

        if self.isnew:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(
                    f,
                    dtype=int,
                    sep=" ",
                    count=nparts * 2
                ).reshape(nparts, 2)[:, 0]
                data = np.ascontiguousarray(data)
            else:
                # unfortunately with gzipped files there does not
                # seem to be an efficient way to load nparts lines
                data = np.zeros(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline().split()[0])

            if self._use_iord:
                data = self._iord_to_fpos[data]
            elif isinstance(self.base, snapshot.ramses.RamsesSnap):
                # AHF only expects three families, DM, star, gas in this order
                # and generates iords on disc according to this rule
                # For classical Ramses snapshots, this is perfectly adequate, but
                # for more modern outputs that have extra tracers, BHs families
                # we need to offset the ids to return the correct slicing
                # TODO These tests on snapshot type might not be necessary
                #  as using the family logic properly should be general.
                #  It is currently kept to ensure 100% backwards compatibility with previous behaviour,
                #  as this code is not explicitly checked by the pynbody test distribution

                if len(self.base) != nd + ns + ng:                      # We have extra families to the base ones
                    # First identify DM, star and gas particles in AHF
                    ahf_dm_mask = data < nd
                    ahf_star_mask = (data >= nd) & (data < nds)
                    ahf_gas_mask = data >= nds

                    # Then offset them by DM family start, to account for
                    # additional families before it, e.g. gas tracers
                    data[np.where(ahf_dm_mask)] += self.base._get_family_slice('dm').start

                    # Star ids used to start at NDM, now they start with the star family slice
                    offset = self.base._get_family_slice('star').start - nd
                    data[np.where(ahf_star_mask)] += offset

                    # Gas ids were greater than NDM + NSTAR, now they start with the gas slice
                    offset = self.base._get_family_slice('gas').start - nds
                    data[np.where(ahf_gas_mask)] += offset

            elif not isinstance(self.base, snapshot.nchilada.NchiladaSnap):
                hi_mask = data >= nds
                data[np.where(hi_mask)] -= nds
                data[np.where(~hi_mask)] += ng
            else:
                st_mask = (data >= nd) & (data < nds)
                g_mask = data >= nds
                data[np.where(st_mask)] += ng
                data[np.where(g_mask)] -= ns
        else:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(f, dtype=int, sep=" ", count=nparts)
            else:
                # see comment above on gzipped files
                data = np.zeros(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline())
        data.sort()
        return data

    def _load_ahf_particles(self, filename):
        if self._use_iord:
            self._iord_to_fpos = np.zeros(self.base['iord'].max()+1,dtype=int)
            self._iord_to_fpos[self.base['iord']] = np.arange(len(self._base()))

        if filename.split("z")[-2][-1] == ".":
            self.isnew = True
        else:
            self.isnew = False

        if self._all_parts is None:
            for h in range(self._nhalos):
                self._halos[h + 1] = DummyHalo()
            return

        with util.open_(filename) as f:
            for h in range(self._nhalos):
                self._halos[h + 1] = Halo(
                    h + 1, self, self.base, self._load_ahf_particle_block(f))
                self._halos[h + 1]._descriptor = "halo_" + str(h + 1)

    def _load_ahf_halos(self, filename):
        # Note: we need to open in 'rt' mode in case the AHF catalogue
        # is gzipped.
        with util.open_(filename, "rt") as f:
            first_line = f.readline()
            lines = f.readlines()

        # get all the property names from the first, commented line
        # remove (#)
        fields = first_line.replace("#", "").split()
        keys = [re.sub(r'\([0-9]*\)', '', field)
                for field in fields]
        # provide translations
        for i, key in enumerate(keys):
            if self.isnew:
                if(key == '#npart'):
                    keys[i] = 'npart'
            else:
                if(key == '#'):
                    keys[i] = 'dumb'
            if(key == 'a'):
                keys[i] = 'a_axis'
            if(key == 'b'):
                keys[i] = 'b_axis'
            if(key == 'c'):
                keys[i] = 'c_axis'
            if(key == 'Mvir'):
                keys[i] = 'mass'

        if self.isnew:
            # fix for column 0 being a non-column in some versions of the AHF
            # output
            if keys[0] == '#':
                keys = keys[1:]

        for h, line in enumerate(lines):
            values = [
                float(x) if any(_ in x for _ in (".", "e", "nan", "inf"))
                else int(x)
                for x in line.split()
            ]
            # XXX Unit issues!  AHF uses distances in Mpc/h, possibly masses as
            # well
            for i, key in enumerate(keys):
                if self.isnew:
                    self._halos[h + 1].properties[key] = values[i]
                else:
                    self._halos[h + 1].properties[key] = values[i - 1]

    def _load_ahf_substructure(self, filename):
        try:
            with util.open_(filename) as f:
                lines = f.readlines()
        except FileNotFoundError:
            self._setup_children()
            return
        logger.info("AHFCatalogue loading substructure")

        # In the substructure catalog, halos are either referenced by their index
        # or by their ID (if they have one).
        ID2index = {}
        for i, halo in self._halos.items():
            # If the "ID" property doesn't exist, use pynbody's internal index
            id = halo.properties.get("ID", i)
            ID2index[id] = i

        for line in lines:
            try:
                haloid, _nsubhalos = (int(x) for x in line.split())
                halo_index = ID2index[haloid]
                children = [
                    ID2index[int(x)] for x in f.readline().split()
                ]
            except ValueError:
                logger.error(
                    "An error occurred while reading substructure file. "
                    "Falling back to using the halo info."
                )
                self._setup_children()
                break
            except KeyError:
                logger.error(
                    (
                        "Could not identify some substructure of "
                        "halo %s. Ignoring"
                    ),
                    haloid + 1
                )
                children = []

            self._halos[halo_index].properties['children'] = children
            for ichild in children:
                self._halos[ichild].properties['parent_id'] = halo_index

    def writegrp(self, grpoutfile=False):
        """
        simply write a skid style .grp file from ahf_particles
        file. header = total number of particles, then each line is
        the halo id for each particle (0 means free).
        """
        snapshot = self[1].ancestor
        try:
            snapshot['grp']
        except Exception:
            self.make_grp()
        if not grpoutfile:
            grpoutfile = snapshot.filename + '.grp'
        logger.info("Writing grp file to %s" % grpoutfile)
        with open(grpoutfile, "w") as fpout:
            print(len(snapshot['grp']), file=fpout)

            # writing 1st to a string sacrifices memory for speed.
            # but this is much faster than numpy.savetxt (could make an option).
            # it is assumed that max halo id <= nhalos (i.e.length of string is set
            # len(str(nhalos))
            stringarray = snapshot['grp'].astype(
                '|S' + str(len(str(self._nhalos))))
            outstring = "\n".join(stringarray)
            print(outstring, file=fpout)

    def writestat(self, snapshot, halos, statoutfile, hubble=None):
        """
        write a condensed skid.stat style ascii file from ahf_halos
        file.  header + 1 halo per line. should reproduce `Alyson's
        idl script' except does not do last 2 columns (Is it a
        satellite?) and (Is central halo is `false'ly split?).  output
        units are set to Mpc Msun, km/s.

        user can specify own hubble constant hubble=(H0/(100
        km/s/Mpc)), ignoring the snaphot arg for hubble constant
        (which sometimes has a large roundoff error).
        """
        s = snapshot
        mindarkmass = min(s.dark['mass'])

        if hubble is None:
            hubble = s.properties['h']

        outfile = statoutfile
        logger.info("Writing stat file to %s" % statoutfile)
        with open(outfile, "w") as fout:
            header = "#Grp  N_tot     N_gas      N_star    N_dark    Mvir(M_sol)       Rvir(kpc)       GasMass(M_sol) StarMass(M_sol)  DarkMass(M_sol)  V_max  R@V_max  VelDisp    Xc   Yc   Zc   VXc   VYc   VZc   Contam   Satellite?   False?   ID_A"
            print(header, file=fpout)
            nhalos = halos._nhalos
            for ii in range(nhalos):
                h = halos[ii + 1].properties  # halo index starts with 1 not 0
                # 'Contaminated'? means multiple dark matter particle masses in halo)"
                icontam = np.where(halos[ii + 1].dark['mass'] > mindarkmass)
                if (len(icontam[0]) > 0):
                    contam = "contam"
                else:
                    contam = "clean"
                # may want to add implement satellite test and false central
                # breakup test.

                n_dark = h['npart'] - h['n_gas'] - h['n_star']
                M_dark = h['mass'] - h['M_gas'] - h['M_star']
                ss = "     "  # can adjust column spacing
                outstring = str(int(h['halo_id'])) + ss
                outstring += str(int(h['npart'])) + ss + str(int(h['n_gas'])) + ss
                outstring += str(int(h['n_star'])) + ss + str(int(n_dark)) + ss
                outstring += str(h['mass'] / hubble) + ss + \
                    str(h['Rvir'] / hubble) + ss
                outstring += str(h['M_gas'] / hubble) + ss + \
                    str(h['M_star'] / hubble) + ss
                outstring += str(M_dark / hubble) + ss
                outstring += str(h['Vmax']) + ss + str(h['Rmax'] / hubble) + ss
                outstring += str(h['sigV']) + ss
                # pos: convert kpc/h to mpc (no h).
                outstring += str(h['Xc'] / hubble / 1000.) + ss
                outstring += str(h['Yc'] / hubble / 1000.) + ss
                outstring += str(h['Zc'] / hubble / 1000.) + ss
                outstring += str(h['VXc']) + ss + \
                    str(h['VYc']) + ss + str(h['VZc']) + ss
                outstring += contam + ss
                outstring += "unknown" + \
                    ss  # unknown means sat. test not implemented.
                outstring += "unknown" + ss  # false central breakup.
                print(outstring, file=fpout)

        return 1

    def writetipsy(self, snapshot, halos, tipsyoutfile, hubble=None):
        """
        write halos to tipsy file (write as stars) from ahf_halos
        file.  returns a shapshot where each halo is a star particle.

        user can specify own hubble constant hubble=(H0/(100
        km/s/Mpc)), ignoring the snaphot arg for hubble constant
        (which sometimes has a large roundoff error).
        """
        import math

        from ..analysis import cosmology
        from ..snapshot import new, tipsy
        s = snapshot
        outfile = tipsyoutfile
        nhalos = halos._nhalos
        nstar = nhalos
        sout = new(star=nstar)  # create new tipsy snapshot written as halos.
        sout.properties['a'] = s.properties['a']
        sout.properties['z'] = s.properties['z']
        sout.properties['boxsize'] = s.properties['boxsize']
        if hubble is None:
            hubble = s.properties['h']
        sout.properties['h'] = hubble
    # ! dangerous -- rho_crit function and unit conversions needs simplifying
        rhocrithhco = cosmology.rho_crit(s, z=0, unit="Msol Mpc^-3 h^2")
        lboxkpc = sout.properties['boxsize'].ratio("kpc a")
        lboxkpch = lboxkpc * sout.properties['h']
        lboxmpch = lboxkpc * sout.properties['h'] / 1000.
        tipsyvunitkms = lboxmpch * 100. / (math.pi * 8. / 3.) ** .5
        tipsymunitmsun = rhocrithhco * lboxmpch ** 3 / sout.properties['h']

        for ii in range(nhalos):
            h = halos[ii + 1].properties
            sout.star[ii]['mass'] = h['mass'] / hubble / tipsymunitmsun
            # tipsy units: box centered at 0. (assume 0<=x<=1)
            sout.star[ii]['x'] = h['Xc'] / lboxkpch - 0.5
            sout.star[ii]['y'] = h['Yc'] / lboxkpch - 0.5
            sout.star[ii]['z'] = h['Zc'] / lboxkpch - 0.5
            sout.star[ii]['vx'] = h['VXc'] / tipsyvunitkms
            sout.star[ii]['vy'] = h['VYc'] / tipsyvunitkms
            sout.star[ii]['vz'] = h['VZc'] / tipsyvunitkms
            sout.star[ii]['eps'] = h['Rvir'] / lboxkpch
            sout.star[ii]['metals'] = 0.
            sout.star[ii]['phi'] = 0.
            sout.star[ii]['tform'] = 0.

        sout.write(fmt=tipsy.TipsySnap, filename=outfile)
        return sout

    def writehalos(self, snapshot, halos, hubble=None, outfile=None):
        """ Write the (ahf) halo catalog to disk.  This is really a
        wrapper that calls writegrp, writetipsy, writestat.  Writes
        .amiga.grp file (ascii group ids), .amiga.stat file (ascii
        halo catalog) and .amiga.gtp file (tipsy halo catalog).
        default outfile base simulation is same as snapshot s.
        function returns simsnap of halo catalog.
        """
        s = snapshot
        grpoutfile = s.filename + ".amiga.grp"
        statoutfile = s.filename + ".amiga.stat"
        tipsyoutfile = s.filename + ".amiga.gtp"
        halos.writegrp(grpoutfile)
        halos.writestat(s, halos, statoutfile, hubble=hubble)
        shalos = halos.writetipsy(s, halos, tipsyoutfile, hubble=hubble)
        return shalos

    @staticmethod
    def _list_possible_candidates(sim, ahf_basename):
        if ahf_basename is not None:
            candidates = glob.glob(f"{ahf_basename}*particles*")
        else:
            candidates = set(glob.glob(f"{sim._filename}*z*particles*"))
            # use a set to ensure that no duplicates can be produced
            # This could arise in an edge case where _filename is a directory
            # and having the "/" at the end of it would lead to a first detection here
            # and a second one again below

            if os.path.isdir(sim._filename):
                candidates = candidates.union(glob.glob(os.path.join(sim._filename, "*z*particles*")))

        return list(candidates)

    @classmethod
    def _can_load(cls, sim, ahf_basename=None, **kwargs):
        candidates = cls._list_possible_candidates(sim, ahf_basename)
        number_ahf_file_candidates = len(candidates)
        return number_ahf_file_candidates > 0

    def _run_ahf(self, sim):
        # if (sim is pynbody.tipsy.TipsySnap) :
        typecode = 90
        # elif (sim is pynbody.gadget.GadgetSnap):
        #   typecode = '60' or '61'
        import pynbody.units as units

        # find AHFstep

        groupfinder = config_parser.get('AHFCatalogue', 'Path')

        if groupfinder == 'None':
            for directory in os.environ["PATH"].split(os.pathsep):
                ahfs = glob.glob(os.path.join(directory, "AHF*"))
                for iahf, ahf in enumerate(ahfs):
                    # if there are more AHF*'s than 1, it's not the last one, and
                    # it's AHFstep, then continue, otherwise it's OK.
                    if ((len(ahfs) > 1) & (iahf != len(ahfs) - 1) &
                            (os.path.basename(ahf) == 'AHFstep')):
                        continue
                    else:
                        groupfinder = ahf
                        break

        if not os.path.exists(groupfinder):
            raise RuntimeError("Path to AHF (%s) is invalid" % groupfinder)

        if (os.path.basename(groupfinder) == 'AHFstep'):
            isAHFstep = True
        else:
            isAHFstep = False
        # build units file
        if isAHFstep:
            with open('tipsy.info', 'w') as f:
                f.write(str(sim.properties['omegaM0']) + "\n")
                f.write(str(sim.properties['omegaL0']) + "\n")
                f.write(str(sim['pos'].units.ratio(
                    units.kpc, a=1) / 1000.0 * sim.properties['h']) + "\n")
                f.write(
                    str(sim['vel'].units.ratio(units.km / units.s, a=1)) + "\n")
                f.write(str(sim['mass'].units.ratio(units.Msol)) + "\n")
                f.close()
                # make input file
                f = open('AHF.in', 'w')
                f.write(sim._filename + " " + str(typecode) + " 1\n")
                f.write(sim._filename + "\n256\n5\n5\n0\n0\n0\n0\n")
        else:
            lgmax = np.min([int(2 ** np.floor(np.log2(
                1.0 / np.min(sim['eps'])))), 131072])
            # hardcoded maximum 131072 might not be necessary

            # make input file
            with open('AHF.in', 'w') as f:
                print(config_parser.get('AHFCatalogue', 'Config', vars={
                    'filename': sim._filename,
                    'typecode': typecode,
                    'gridmax': lgmax
                }), file=f)

                print(config_parser.get('AHFCatalogue', 'ConfigTipsy', vars={
                    'omega0': sim.properties['omegaM0'],
                    'lambda0': sim.properties['omegaL0'],
                    'boxsize': sim['pos'].units.ratio('Mpc a h^-1', **sim.conversion_context()),
                    'vunit': sim['vel'].units.ratio('km s^-1 a', **sim.conversion_context()),
                    'munit': sim['mass'].units.ratio('Msol h^-1', **sim.conversion_context()),
                    'eunit': 0.03  # surely this can't be right?
                }), file=f)

        if not os.path.exists(sim._filename):
            os.system("gunzip " + sim._filename + ".gz")
        # determine parallel possibilities

        if os.path.exists(groupfinder):
            # run it
            os.system(groupfinder + " AHF.in")
            return

    @staticmethod
    def _can_run(sim):
        if config_parser.getboolean('AHFCatalogue', 'AutoRun'):
            if config_parser.get('AHFCatalogue', 'Path') == 'None':
                for directory in os.environ["PATH"].split(os.pathsep):
                    if (len(glob.glob(os.path.join(directory, "AHF*"))) > 0):
                        return True
            else:
                path = config_parser.get('AHFCatalogue', 'Path')
                return os.path.exists(path)
        return False
